/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "wstream_chan.hh"
#include "wcondition.hh"
#include "weventlist.hh"
#include "wtransform.hh"
#include "woutput.hh"

#include "matlab_fcs.hh"
#include "MultiDacc.hh"
#include "lcl_array.hh"

using namespace wpipe;

//======================================  Construct a channel streaming block
wstream_chan::wstream_chan(const wparameters& par, int dbugLvl) 
   : pars(par), debugLevel(dbugLvl), minimumTimeShift(0), maximumTimeShift(0)
{

   //////////////////////////////////////////////////////////////////////////
   //                           tile signal space                          //
   //////////////////////////////////////////////////////////////////////////
   wlog(debugLevel, 1, "tiling search space");

   // generate Q transform tiling
   double fSample = pars.sampleFrequency;
   tiling.init(pars.blockDuration, pars.qRange, pars.frequencyRange, 
	       fSample, pars.maximumMismatch, 
	       pars.highPassCutoff, pars.lowPassCutoff, 
	       pars.whiteningDuration, pars.transientFactor);

   if (debugLevel >=2) {
      cout << "  whiteningDuration:   " << tiling.whiteningDuration() << endl;
      cout << "  transientDuration:   " << tiling.transientDuration() << endl;
   }

   /////////////////////////////////////////////////////////////////////////////
   //                              partition segment                          //
   /////////////////////////////////////////////////////////////////////////////
   wlog(debugLevel, 1, "portioning segment");

   //----------------------------------  get max, min time shift (sample
   //                                    alignment forced in pars)
   if (!pars.timeShifts.empty()) {
      maximumTimeShift = pars.timeShifts[0];
      minimumTimeShift = pars.timeShifts[0];
      size_t N = pars.timeShifts.size();
      for (size_t i=1; i<N; i++) {
	 if (pars.timeShifts[i] > maximumTimeShift) {
	    maximumTimeShift = pars.timeShifts[i];
	 } else if (pars.timeShifts[i] < minimumTimeShift) {
	    minimumTimeShift = pars.timeShifts[i];
	 }
      }
   }
}

//======================================  wstream_chan destructor.
wstream_chan::~wstream_chan(void) {
}

//======================================  Request all useful channels.
void
wstream_chan::addChannels(MultiDacc& md) const {

   //-----------------------------------  Add data channels
   size_t nChannel = pars.channelNames.size();
   for (size_t i=0; i<nChannel; ++i) {
      md.addChannel(pars.channelNames[i].c_str(), 0);
   }

   //-----------------------------------  Add state channels
   size_t nStates  = pars.stateNames.size();
   for (size_t i=0; i<nStates; ++i) {
      md.addChannel(pars.stateNames[i].c_str(), 0);
   }

   //-----------------------------------  Add request channels
   size_t nInject  = pars.injectionNames.size();
   for (size_t i=0; i < nInject; ++i) {
      const string& name = pars.injectionNames[i];
      if (!name.empty() && name != "NONE"){
	 md.addChannel(name.c_str(), 0);
      }
   }
}

//======================================  Add output files.
void
wstream_chan::addOutputFiles(woutput& wo, const std::string& fileTime, 
			     const std::string& outdir) const {
   wo.addMany(pars.channelNames, fileTime, outdir, 
	      pars.triggerFormat, pars.triggerFiles);
}

//======================================  Read injection data
void
wstream_chan::inject(MultiDacc& md) {
   size_t nInject = pars.injectionChannels();
   if (!nInject) return;

   wlog(debugLevel, 1, "  reading injection data");

   //-----------------------------------  Veto missing, inconsistent injections
   size_t nChan = pars.numberOfChannels();
   for (size_t chan=0; chan<nChan; chan++) {
      string& name = pars.injectionNames[chan];
      if (name.empty()) continue;

      TSeries injectionData = *md.refData(name.c_str());

      if (injectionData.empty()) {
	 error("error reading injection data");
      }

      if (injectionData.getTStep() != input_data[chan].getTStep()) {
	 error("inconsistent detector and injection data");
      }

      // Scale injection data and add it in.
      injectionData    *= pars.injectionFactors[chan];
      input_data[chan] += injectionData;

      // report injection data loaded
      if (debugLevel >= 1) {
	 cout <<  "    " << name << ":        loaded x " 
	      << pars.injectionFactors[chan] << endl;
      }
   }  // end of loop over channels.
}

//======================================  Calculate minimum overlap
double 
wstream_chan::minimumOverlap(void) const {

   //-----------------------------------  determine minimum block overlap
   double minimumBlockOverlap = 2 * tiling.transientDuration() + 
                                pars.extraBlockOverlap;

   if (debugLevel >=2) {
      cout << "  minimumBlockOverlap: " << minimumBlockOverlap << endl;
   }
      
   //-----------------------------------  Segment livetime lost to time shifts
   //double timeShiftLoss = maximumTimeShift - minimumTimeShift;

   //-----------------------------------  Round overlap up to nearest second
   double blockOverlap = floor(minimumBlockOverlap);
   if (blockOverlap < minimumBlockOverlap) blockOverlap += 1.0;
   return blockOverlap;
}

//======================================  Perform processing
void 
wstream_chan::process(const woutput& outputFiles, const Time& blockStartTime, 
		      Interval blockDuration) {
   size_t   nChans       = pars.numberOfChannels();
   str_vect channelNames = pars.channelNames;
   double   fSample      = pars.sampleFrequency;

   dble_vect coordinate;

#ifndef NO_COHERENT
   //////////////////////////////////////////////////////////////////////
   //                    transform sky coordinate                      //
   //////////////////////////////////////////////////////////////////////

   // if (the skyPosition parameter is not empty, determine coordinate
   if (!pars.skyPosition.empty()) {

      // gps center time of block
      Time blockCenterTime = blockStartTime + (blockDuration / 2);
      if (debugLevel >= 2) {
	 cout << "  block center time:       " << blockCenterTime << endl;
      }

      // convert sky coordinate into geocentric coordinate
      wconvertskycoordinates coordinate(pars.skyPosition, blockCenterTime, 
					pars.skyCoordinateSystem, "geocentric");

      if (debugLevel >= 1) {
	 cout <<  "  sky coordinate:          [" << coordinate[1] << " " 
	      << coordinate[2] << endl;
      }
   }  // if (!skyposition.empty())
#endif  

   //-----------------------------------  pre-process data
   tser_vect data(nChans);
   string networkString;
   for (size_t chan=0; chan < nChans; chan++) {
      if (!validChannels[chan]) continue;

      data[chan]=input_data[chan].extract(blockStartTime, blockDuration);
      if (debugLevel) {
	 cerr << "Channel " << channelNames[chan] << " start: " 
	      << data[chan].getStartTime() << " end: "
	      << data[chan].getEndTime() << endl;
      }

      // test for missing detector data
      if (data[chan].empty()) error("error reading detector data");

      // standardize channel names
      channelNames[chan]=strrep(channelNames[chan], ";", ":");

      //--------------------------------  record analyzed network
      if (chan != 0) networkString += ",";
      networkString += channelNames[chan].substr(0,2);
   }

   //-----------------------------------  resample data
   wlog(debugLevel, 1, "  resampling data");
   tser_vect resdata = resample.wresample(data, fSample);
   if (debugLevel) {
      cerr << "Resampled channel " << channelNames[0] << " start: " 
	   << resdata[0].getStartTime() << " end: "
	   << resdata[0].getEndTime() << endl;
   }

   //------------------------------------  Condition data
   wlog(debugLevel, 1, "  conditioning data");
   wcondition conditioner(resdata, tiling, pars.doubleWhiten);

   //-----------------------------------  Transform and threshold
   Time thresholdReferenceTime = blockStartTime + pars.blockDuration*0.5;
   dble_vect thresholdTimeRange;
   dble_vect thresholdFrequencyRange;
   dble_vect thresholdQRange;

   // The original version combined the transformation and triggering.
   // I don't understand why this would be necessary except for memory 
   // management. this could be done in c++ but it certainly isn't a 
   // critical problem. If it should prove to be important at a later 
   // time, the channel-by-channel transform+thresholding can be 
   // implemented using the single channel transform and thresholding 
   // classes.
   wlog(debugLevel, 1, "  transforming data");
   dft_vect condDft, coeffs;
   conditioner.whitenedDFT(condDft);
   conditioner.coefficientDFT(coeffs);

   wtransform transforms(condDft, tiling, pars.outlierFactor, 
			 pars.analysisMode, channelNames, coeffs, 
			 coordinate);

   wlog(debugLevel, 1, "  thresholding data");
   weventstack triggers, triggersThresholded, triggersDownselected;
   triggersThresholded.wthreshold(transforms, tiling, blockStartTime, 
				  pars.falseEventRate, 
				  thresholdReferenceTime, 
				  thresholdTimeRange, 
				  thresholdFrequencyRange, 
				  thresholdQRange, 
				  pars.maximumSignificants, 
				  pars.analysisMode, pars.falseVetoRate, 
				  pars.uncertaintyFactor, 
				  pars.correlationFactor, debugLevel);

   // report significant tile numbers
   if (debugLevel >= 1) {
      triggersThresholded.status(cout);
   }

   //-----------------------------------  select significant triggers
   wlog(debugLevel, 1, "  selecting significant triggers");

   // excluding overlapping significant tiles
   triggersDownselected.wselect(triggersThresholded, 
				pars.durationInflation,
				pars.bandwidthInflation, 
				pars.maximumTriggers, 
				debugLevel);

   // replace triggers
   triggers = triggersDownselected;

   // report downselected tile numbers
   if (debugLevel >= 1) {
      triggers.status(cout);
   }

   //-----------------------------------  write single detector triggers
   if (outputFiles.enabled("DOWNSELECT")) {
      wlog(debugLevel, 1, "  writing selected triggers");
      triggers.writeEvents(outputFiles["DOWNSELECT"], 
			   pars.triggerFields, 
			   pars.triggerFormat);
   }

   //-----------------------------------  Veto inconsistent triggers
#ifndef NO_COHERENT
   // if veto is requested and analysis mode is coherent
   if (pars.applyVeto && pars.analysisMode == "coherent") {
      wlog(debugLevel, 1, "  applying veto");

      //-----------------------------  Apply null stream veto
      eventstack triggersVetoed;
      triggersVetoed.wveto(triggers, pars.durationInflation, 
			   pars.bandwidthInflation, 
			   vetoDurationFactor, vetoBandwidthFactor, 
			   maximumConsistents, debugLevel);

      // replace triggers
      triggers = triggersVetoed;

      // report consistent tile numbers
      if (debugLevel >= 1) {
	 triggersVetoed.status(cout);
      }
   } // end test for apply veto
#endif // !defined(NO_COHERENT)

   //-----------------------------------  Cluster triggers

   if (pars.applyClustering || pars.bayesian) {
      wlog(debugLevel, 1, "  applying clustering");

      weventstack triggersClustered = triggers;
      weventstack clusters;

      //---------------------------------  Density clustering
      string method = toupper(pars.clusterMethod);
      if (pars.clusterMethod == "density") {
	 clusters.wcluster(triggersClustered, pars.clusterRadius, 
			   pars.clusterDensity, pars.clusterSingles, 
			   pars.distanceMetric, pars.durationInflation, 
			   pars.bandwidthInflation, debugLevel);
      }

      //-----------------------------  Hierarchical clustering
      else if (pars.clusterMethod == "hierarchical") {
	 clusters.wcluster(triggersClustered, pars.clusterLinkage, 
			   pars.clusterCriterion, pars.clusterThreshold, 
			   pars.distanceMetric, pars.durationInflation,
			   pars.bandwidthInflation, debugLevel);
      }

      //--------------------------------  Invalid clustering method
      else {
	 error("Unknown cluster method");
      }

      //--------------------------------  Replace triggers
      triggers = triggersClustered;

      //--------------------------------  Report cluster numbers
      if (debugLevel >= 1) {
	 clusters.status(cout);
      }

      //--------------------------------  Write cluster triggers
      if (outputFiles.enabled("CLUSTER")) {
	 wlog(debugLevel, 1, "  writing cluster triggers");
	 if (pars.writeClusters) {
	    clusters.writeEvents(outputFiles["CLUSTER"], 
				 pars.triggerFields,
				 pars.triggerFormat);
	 } 
	 else {
	    triggersClustered.writeEvents(outputFiles["CLUSTER"], 
					  pars.triggerFields, 
					  pars.triggerFormat);
	 }
      }
   } // end test for apply clustering

   //-----------------------------------  Coincident triggers
#ifndef NO_COINCIDENCE
   bool coincidentsFound = false;

   //-----------------------------------  apply coincidence if requested
   if ((pars.coincidenceNumber > 1) && (numberOfChannels > 1)) {
      wlog(debugLevel, 1, "  coinciding triggers");
	 
      // apply coincidence
      wcoincide triggersCoincident(triggers, 
				   pars.coincidenceNumber, 
				   pars.durationInflation,
				   pars.bandwidthInflation, 
				   pars.maximumCoincidents, 
				   debugLevel);

      // replace triggers
      triggers = triggersCoincident;
  
      // cull coincident clusters based on coincident triggers
      if (exist("clusters","var")) {
	 wcoincidecluster clustersCoincident(triggersCoincident, 
					     clusters, debugLevel);
      
	 // replace clusters
	 clusters = clustersCoincident;
      }

      //--------------------------------  Note if any coincidents were found
      if (!triggersCoincident.empty()) {
	 coincidentsFound = true;
      }

      //--------------------------------  Report coincidence numbers
      if (debugLevel >= 1) {
	 for (size_t chanNumber=0; chanNumber < triggersCoincident.size();
	      chanNumber++) {
	    cout <<  "    " << triggersCoincident[chanNumber].channelName 
		 << ": " << triggersCoincident[chanNumber].size() 
		 << " coincidents" << endl;
	 }
      }

      //--------------------------------  Write coincident triggers
      if (outputFiles.enabled("COINCIDE")) {
	 wlog(debugLevel, 1, "  writing coincident triggers");

	 // write triggers or clusters
	 if (pars.applyClustering && pars.writeClusters) {
	    wwriteevents(clustersCoincident, outputFiles.COINCIDE, 
			 pars.triggerFields, pars.triggerFormat);
	 }
	 else {
	    wwriteevents(triggersCoincident, outputFiles.COINCIDE, 
			 pars.triggerFields, pars.triggerFormat);
	 }
      }
   }    // end test for apply coincidence

   ///////////////////////////////////////////////////////////////////
   //                    EVENT FOLLOWUP
   ///////////////////////////////////////////////////////////////////

   // apply followup if requested and coincident tiles 
   // and clusters have been found
   if (coincidentsFound) {

      //--------------------------------  Create the output event structure
      event.id = "Discrete Q-transform event structure";
      event.network = networkString;
      event.blockStartTime = blockStartTime;
      event.blockStopTime = blockStopTime;
      event.livetime = pars.blockDuration - 2*tiling.transientDuration;

      event.outputFields = {"network", "blockStartTime",
			    "blockStopTime", "livetime"};

      //--------------------------------  Find candidate event cluster
      wlog(debugLevel, 1, "  finding candidate cluster");

      wfindcandidate candidateCluster(clustersCoincident, 
				      triggersClustered,
				      eventTime, pars, debugLevel);

      event.time = candidateCluster.clusterTime;
      event.frequency = candidateCluster.clusterFrequency;
      event.duration = candidateCluster.clusterDuration;
      event.bandwidth = candidateCluster.clusterBandwidth;

      event.outputFields = {event.outputFields{:}, "time", 
			    "frequency", "duration", "bandwidth"};

      ////////////////////////////////////////////////////////////////////////
      //                     Follow-up Analyses                             //
      ////////////////////////////////////////////////////////////////////////
#ifndef NO_FOLLOWUP
      if (pars.bayesian || pars.xCoherentCheck || pars.maxEnt){
	 wlog(debugLevel, 1, "  Event followups:");
	    
	 [event, skymap] = 
	    wfollowups(event, data, coefficients, tiling, 
		       blockStartTime, coordinate, candidateCluster, 
		       channelNames, pars, outputFiles, 
		       debugLevel);
      }
#endif // !defined( NO_FOLLOWUP)

      //--------------------------------  Write event
      wlog(debugLevel, 1, "  writing event");
      wwriteevents(event, outputFiles.EVENTS, event.outputFields, 
		   pars.triggerFormat);
   } // end if (coincidentsFound)
#endif // !defined(NO_COINCIDENCE)
}

//======================================  Read data and state information.
void 
wstream_chan::read_data(MultiDacc& md, const Time& startTime, double duration) {
   size_t nChans = pars.numberOfChannels();
   if (!nChans) return;
   if (validChannels.size() < nChans) {
      error("validChannels not initialized");
   }

   //----------------------------------- Check the number of state channels
   size_t nStateChans = pars.stateNames.size();
   if (nStateChans > nChans) {
      error("Invalid number of state channels");
   }

   //-----------------------------------  Evaluate state data
   for (size_t i=0; i<nStateChans; i++) {
      string sname = pars.stateNames[i];
      int mask = int(pars.stateMasks[i]);
      if (!mask || sname.empty() || sname == "NONE") continue;

      //--------------------------------  See if state is already evaluated
      bool sameas = false;
      for (size_t j=0; j<i; ++j) {
	 if (sname == pars.stateNames[j] && mask == pars.stateMasks[j]) {
	    validChannels[i] = validChannels[j];
	    sameas = true;
	    break;
	 }
      }
      if (sameas) continue;

      //--------------------------------  Quick check for missing or invalid 
      //                                  state data
      TSeries stateData=md.refData(sname.c_str())->extract(startTime, duration);
      if (stateData.empty() || stateData.getNLess(double(mask))) {
	 if (debugLevel >= 1) {
	    cout <<  "    " << sname << ":        data fails quick test "
		 << "(state < " << mask << ")" << endl; 
	 }
	 if (pars.errorOnStateError && stateData.getMinimum() < 0) {
	    error("Error reading detector state");
	 }
	 validChannels[i] = false;
      }

      //--------------------------------  Loop over all state data and test
      //                                  that mask bits are on.
      else {
	 int N = stateData.getNSample();
	 lcl_array<int> state_chan(N);
	 stateData.getData(N, state_chan);
	 for (int i=0; i<N; i++) {
	    if ((state_chan[i] & mask) != mask) {
	       if (debugLevel >= 1) {
		  cout <<  "    " << sname 
		       << ":        detector not in requested state" << endl; 
	       }
	       validChannels[i] = false;
	       break;
	    }
	 }
      }
   }

   //-----------------------------------  Read in the data
   // Time fillTime = md.getFillTime();
   for (size_t i=0; i<nChans; i++) {
      //--------------------------------  Don't bother if data are uninteresting
      if (!validChannels[i]) continue;
      TSeries* ts = md.refData(pars.channelNames[i].c_str());
      if (!ts) {
	 cerr << "Data not available for channel: " << pars.channelNames[i]
	      << endl;
	 md.list(cerr);
	 throw runtime_error("wchunk: Error reading data");
      }
      Time tStart = input_data[i].getStartTime();
      if (!input_data[i].empty() && tStart < startTime) {
	 Interval tDelete = startTime - tStart;
	 input_data[i].eraseStart(tDelete);
      }

      //--------------------------------  Scale data and append to input buffer
      double dFact = pars.dataFactors[i];
      if (input_data[i].empty() || 
	  ts->getStartTime() > input_data[i].getEndTime()) {
	 input_data[i] = *ts;
	 if (dFact != 1.0) input_data[i] *= dFact;
      } else if (dFact != 1.0) {
	 TSeries scaled_ts = *ts;
	 scaled_ts *= dFact;
	 input_data[i].Append(scaled_ts);
      } else {
	 input_data[i].Append(*ts);
      }

      //--------------------------------  Report data loaded
      if (debugLevel >= 1) {
	 cout <<  "    " << pars.channelNames[i] << ":        loaded x "
	      <<  dFact << endl;
      }
   }
}

//======================================  Reset, ready for start of next chunk
void
wstream_chan::start_chunk(void) {
   validChannels.clear();
   size_t nChan = pars.numberOfChannels();
   validChannels.resize(nChan, true);
   if (input_data.size() != nChan) input_data.resize(nChan);
}

//======================================  Reset, ready for start of next chunk
void 
wstream_chan::validChans(str_vect& processedChannels) const {
   size_t nChan = pars.numberOfChannels();
   for (size_t i=0; i<nChan; i++) {
      if (validChannels[i]) processedChannels.push_back(pars.channelNames[i]);
   }
}
